column_device_view.cuh
Go to the documentation of this file.
1 /*
2  * SPDX-FileCopyrightText: Copyright (c) 2019-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
3  * SPDX-License-Identifier: Apache-2.0
4  */
5 #pragma once
6 
9 #include <cudf/detail/utilities/alignment.hpp>
10 #include <cudf/lists/list_view.hpp>
15 #include <cudf/utilities/span.hpp>
17 
18 #include <rmm/cuda_stream_view.hpp>
19 #include <rmm/resource_ref.hpp>
20 
21 #include <cuda/iterator>
22 #include <cuda/std/utility>
23 #include <thrust/iterator/transform_iterator.h>
24 
25 #include <functional>
26 
32 namespace CUDF_EXPORT cudf {
33 
40 class alignas(16) column_device_view : public column_device_view_core {
41  public:
43 
44  column_device_view() = delete;
45  ~column_device_view() = default;
60 
70  column_device_view(column_view column, void* h_ptr, void* d_ptr);
71 
89  size_type size) const noexcept
90  {
91  return column_device_view{this->type(),
92  size,
93  this->head(),
94  this->null_count(),
95  this->null_mask(),
96  this->offset() + offset,
97  static_cast<column_device_view*>(_children),
98  this->num_child_columns()};
99  }
100 
118  template <typename T, CUDF_ENABLE_IF(is_rep_layout_compatible<T>())>
119  [[nodiscard]] __device__ T element(size_type element_index) const noexcept
120  {
121  return base::element<T>(element_index);
122  }
123 
135  template <typename T, CUDF_ENABLE_IF(cuda::std::is_same_v<T, string_view>)>
136  [[nodiscard]] __device__ T element(size_type element_index) const noexcept
137  {
138  return base::element<T>(element_index);
139  }
140 
151  template <typename T, CUDF_ENABLE_IF(cudf::is_fixed_point<T>())>
152  [[nodiscard]] __device__ T element(size_type element_index) const noexcept
153  {
154  return base::element<T>(element_index);
155  }
156 
157  private:
163  struct index_element_fn {
164  template <typename IndexType,
165  CUDF_ENABLE_IF(is_index_type<IndexType>() and cuda::std::is_signed_v<IndexType>)>
166  __device__ size_type operator()(column_device_view const& indices, size_type index)
167  {
168  return static_cast<size_type>(indices.element<IndexType>(index));
169  }
170 
171  template <typename IndexType,
172  typename... Args,
173  CUDF_ENABLE_IF(not(is_index_type<IndexType>() and cuda::std::is_signed_v<IndexType>))>
174  __device__ size_type operator()(Args&&... args)
175  {
176  CUDF_UNREACHABLE("dictionary indices must be a signed integral type");
177  }
178  };
179 
180  public:
205  template <typename T, CUDF_ENABLE_IF(cuda::std::is_same_v<T, dictionary32>)>
206  [[nodiscard]] __device__ T element(size_type element_index) const noexcept
207  {
208  size_type index = element_index + offset(); // account for this view's _offset
209  auto const indices = child(0);
210  return dictionary32{type_dispatcher(indices.type(), index_element_fn{}, indices, index)};
211  }
212 
219  template <typename T>
220  CUDF_HOST_DEVICE static constexpr bool has_element_accessor()
221  {
222  return has_element_accessor_impl<column_device_view, T>::value;
223  }
224 
226  using count_it = cuda::counting_iterator<size_type>;
230  template <typename T>
231  using const_iterator = thrust::transform_iterator<detail::value_accessor<T>, count_it>;
232 
248  template <typename T, CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
249  [[nodiscard]] const_iterator<T> begin() const
250  {
252  }
253 
268  template <typename T, CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
269  [[nodiscard]] const_iterator<T> end() const
270  {
271  return const_iterator<T>{count_it{size()}, detail::value_accessor<T>{*this}};
272  }
273 
277  template <typename T, typename Nullate>
279  thrust::transform_iterator<detail::optional_accessor<T, Nullate>, count_it>;
280 
284  template <typename T, bool has_nulls>
286  thrust::transform_iterator<detail::pair_accessor<T, has_nulls>, count_it>;
287 
293  template <typename T, bool has_nulls>
295  thrust::transform_iterator<detail::pair_rep_accessor<T, has_nulls>, count_it>;
296 
351  template <typename T,
352  typename Nullate,
353  CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
354  auto optional_begin(Nullate has_nulls) const
355  {
358  }
359 
381  template <typename T,
382  bool has_nulls,
383  CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
385  {
388  }
389 
413  template <typename T,
414  bool has_nulls,
415  CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
417  {
420  }
421 
438  template <typename T,
439  typename Nullate,
440  CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
441  auto optional_end(Nullate has_nulls) const
442  {
445  }
446 
458  template <typename T,
459  bool has_nulls,
460  CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
462  {
465  }
466 
479  template <typename T,
480  bool has_nulls,
481  CUDF_ENABLE_IF(column_device_view::has_element_accessor<T>())>
483  {
486  }
487 
507  static std::unique_ptr<column_device_view, std::function<void(column_device_view*)>> create(
508  column_view source_view,
511 
518  void destroy();
519 
527  static std::size_t extent(column_view const& source_view);
528 
535  [[nodiscard]] __device__ column_device_view child(size_type child_index) const noexcept
536  {
537  return static_cast<column_device_view*>(_children)[child_index];
538  }
539 
545  [[nodiscard]] __device__ device_span<column_device_view const> children() const noexcept
546  {
547  return {static_cast<column_device_view*>(_children), static_cast<std::size_t>(_num_children)};
548  }
549 
555  [[nodiscard]] CUDF_HOST_DEVICE size_type num_child_columns() const noexcept
556  {
557  return _num_children;
558  }
559 
560  private:
575  size_type size,
576  void const* data,
578  bitmask_type const* null_mask,
579  size_type offset,
580  column_device_view* children,
581  size_type num_children)
583  type, size, data, null_count, null_mask, offset, children, num_children}
584  {
585  }
586 
597  column_device_view(column_view source);
598 };
599 
607  public:
609 
610  mutable_column_device_view() = delete;
611  ~mutable_column_device_view() = default;
626 
637 
657  static std::unique_ptr<mutable_column_device_view,
658  std::function<void(mutable_column_device_view*)>>
662 
680  template <typename T, CUDF_ENABLE_IF(is_rep_layout_compatible<T>())>
681  [[nodiscard]] __device__ T& element(size_type element_index) const noexcept
682  {
683  return base::element<T>(element_index);
684  }
685 
692  template <typename T>
693  CUDF_HOST_DEVICE static constexpr bool has_element_accessor()
694  {
695  return has_element_accessor_impl<mutable_column_device_view, T>::value;
696  }
697 
699  using count_it = cuda::counting_iterator<size_type>;
703  template <typename T>
704  using iterator = thrust::transform_iterator<detail::mutable_value_accessor<T>, count_it>;
705 
716  template <typename T, CUDF_ENABLE_IF(mutable_column_device_view::has_element_accessor<T>())>
718  {
720  }
721 
732  template <typename T, CUDF_ENABLE_IF(mutable_column_device_view::has_element_accessor<T>())>
734  {
736  }
737 
744  [[nodiscard]] __device__ mutable_column_device_view child(size_type child_index) const noexcept
745  {
746  return static_cast<mutable_column_device_view*>(_children)[child_index];
747  }
748 
757  static std::size_t extent(mutable_column_view source_view);
758 
773  [[nodiscard]] static auto create(data_type type,
774  size_type size,
775  void const* data,
776  bitmask_type const* null_mask,
777  size_type offset,
778  mutable_column_device_view* children,
779  size_type num_children)
780  {
781  return mutable_column_device_view{type, size, data, null_mask, offset, children, num_children};
782  }
783 
790  void destroy();
791 
792  private:
802 
816  size_type size,
817  void const* data,
818  bitmask_type const* null_mask,
819  size_type offset,
820  mutable_column_device_view* children,
821  size_type num_children)
822  : mutable_column_device_view_core{type, size, data, null_mask, offset, children, num_children}
823  {
824  }
825 };
826 
827 namespace detail {
828 
843 template <typename T>
846 
852  value_accessor(column_device_view const& _col) : col{_col}
853  {
854  CUDF_EXPECTS(type_id_matches_device_storage_type<T>(col.type().id()), "the data type mismatch");
855  }
856 
862  __device__ T operator()(cudf::size_type i) const { return col.element<T>(i); }
863 };
864 
891 template <typename T, typename Nullate>
894 
901  optional_accessor(column_device_view const& _col, Nullate with_nulls)
902  : col{_col}, has_nulls{with_nulls}
903  {
904  CUDF_EXPECTS(type_id_matches_device_storage_type<T>(col.type().id()), "the data type mismatch");
905  if (with_nulls) { CUDF_EXPECTS(_col.nullable(), "Unexpected non-nullable column."); }
906  }
907 
915  __device__ inline cuda::std::optional<T> operator()(cudf::size_type i) const
916  {
917  if (has_nulls) {
918  return (col.is_valid_nocheck(i)) ? cuda::std::optional<T>{col.element<T>(i)}
919  : cuda::std::optional<T>{cuda::std::nullopt};
920  }
921  return cuda::std::optional<T>{col.element<T>(i)};
922  }
923 
924  Nullate has_nulls{};
925 };
926 
946 template <typename T, bool has_nulls = false>
949 
955  pair_accessor(column_device_view const& _col) : col{_col}
956  {
957  CUDF_EXPECTS(type_id_matches_device_storage_type<T>(col.type().id()), "the data type mismatch");
958  if (has_nulls) { CUDF_EXPECTS(_col.nullable(), "Unexpected non-nullable column."); }
959  }
960 
967  __device__ inline cuda::std::pair<T, bool> operator()(cudf::size_type i) const
968  {
969  return {col.element<T>(i), (has_nulls ? col.is_valid_nocheck(i) : true)};
970  }
971 };
972 
992 template <typename T, bool has_nulls = false>
995 
997 
1003  pair_rep_accessor(column_device_view const& _col) : col{_col}
1004  {
1005  CUDF_EXPECTS(type_id_matches_device_storage_type<T>(col.type().id()), "the data type mismatch");
1006  if (has_nulls) { CUDF_EXPECTS(_col.nullable(), "Unexpected non-nullable column."); }
1007  }
1008 
1015  __device__ inline cuda::std::pair<rep_type, bool> operator()(cudf::size_type i) const
1016  {
1017  return {get_rep<T>(i), (has_nulls ? col.is_valid_nocheck(i) : true)};
1018  }
1019 
1020  private:
1021  template <typename R>
1022  [[nodiscard]] __device__ inline auto get_rep(cudf::size_type i) const
1023  requires(std::is_same_v<R, rep_type>)
1024  {
1025  return col.element<R>(i);
1026  }
1027 
1028  template <typename R>
1029  [[nodiscard]] __device__ inline auto get_rep(cudf::size_type i) const
1030  requires(not std::is_same_v<R, rep_type>)
1031  {
1032  return col.element<R>(i).value();
1033  }
1034 };
1035 
1047 template <typename T>
1050 
1057  {
1058  CUDF_EXPECTS(type_id_matches_device_storage_type<T>(col.type().id()), "the data type mismatch");
1059  }
1060 
1067  __device__ T& operator()(cudf::size_type i) { return col.element<T>(i); }
1068 };
1069 
1095 template <typename ColumnDeviceView, typename ColumnViewIterator>
1096 ColumnDeviceView* child_columns_to_device_array(ColumnViewIterator child_begin,
1097  ColumnViewIterator child_end,
1098  void* h_ptr,
1099  void* d_ptr)
1100 {
1101  ColumnDeviceView* d_children = detail::align_ptr_for_type<ColumnDeviceView>(d_ptr);
1102  auto num_children = std::distance(child_begin, child_end);
1103  if (num_children > 0) {
1104  // The beginning of the memory must be the fixed-sized ColumnDeviceView
1105  // struct objects in order for d_children to be used as an array.
1106  auto h_column = detail::align_ptr_for_type<ColumnDeviceView>(h_ptr);
1107  auto d_column = d_children;
1108 
1109  // Any child data is assigned past the end of this array: h_end and d_end.
1110  auto h_end = reinterpret_cast<int8_t*>(h_column + num_children);
1111  auto d_end = reinterpret_cast<int8_t*>(d_column + num_children);
1112  std::for_each(child_begin, child_end, [&](auto const& col) {
1113  // inplace-new each child into host memory
1114  new (h_column) ColumnDeviceView(col, h_end, d_end);
1115  h_column++; // advance to next child
1116  // update the pointers for holding this child column's child data
1117  auto col_child_data_size = ColumnDeviceView::extent(col) - sizeof(ColumnDeviceView);
1118  h_end += col_child_data_size;
1119  d_end += col_child_data_size;
1120  });
1121  }
1122  return d_children;
1123 }
1124 
1125 } // namespace detail
1126 } // namespace CUDF_EXPORT cudf
An immutable, non-owning view of device data as a column of elements that is trivially copyable and u...
An immutable, non-owning view of device data as a column of elements that is trivially copyable and u...
column_device_view child(size_type child_index) const noexcept
Returns the specified child.
void destroy()
Destroy the column_device_view object.
column_device_view & operator=(column_device_view &&)=default
Move assignment operator.
static constexpr CUDF_HOST_DEVICE bool has_element_accessor()
For a given T, indicates if column_device_view::element<T>() has a valid overload.
const_pair_iterator< T, has_nulls > pair_end() const
Return a pair iterator to the element following the last element of the column.
const_pair_rep_iterator< T, has_nulls > pair_rep_end() const
Return a pair iterator to the element following the last element of the column.
const_pair_iterator< T, has_nulls > pair_begin() const
Return a pair iterator to the first element of the column.
T element(size_type element_index) const noexcept
Returns a copy of the element at the specified index.
static std::unique_ptr< column_device_view, std::function< void(column_device_view *)> > create(column_view source_view, rmm::cuda_stream_view stream=cudf::get_default_stream(), rmm::device_async_resource_ref mr=cudf::get_current_device_resource_ref())
Factory to construct a column view that is usable in device memory.
cuda::counting_iterator< size_type > count_it
Counting iterator.
column_device_view & operator=(column_device_view const &)=default
Copy assignment operator.
CUDF_HOST_DEVICE column_device_view slice(size_type offset, size_type size) const noexcept
Get a new column_device_view which is a slice of this column.
thrust::transform_iterator< detail::value_accessor< T >, count_it > const_iterator
Iterator for navigating this column.
auto optional_end(Nullate has_nulls) const
Return an optional iterator to the element following the last element of the column.
thrust::transform_iterator< detail::pair_accessor< T, has_nulls >, count_it > const_pair_iterator
Pair iterator for navigating this column.
const_pair_rep_iterator< T, has_nulls > pair_rep_begin() const
Return a pair iterator to the first element of the column.
thrust::transform_iterator< detail::optional_accessor< T, Nullate >, count_it > const_optional_iterator
Optional iterator for navigating this column.
CUDF_HOST_DEVICE size_type num_child_columns() const noexcept
Returns the number of child columns.
auto optional_begin(Nullate has_nulls) const
Return an optional iterator to the first element of the column.
column_device_view(column_device_view const &)=default
Copy constructor.
column_device_view(column_device_view &&)=default
Move constructor.
const_iterator< T > begin() const
Return an iterator to the first element of the column.
thrust::transform_iterator< detail::pair_rep_accessor< T, has_nulls >, count_it > const_pair_rep_iterator
Pair rep iterator for navigating this column.
static std::size_t extent(column_view const &source_view)
Return the size in bytes of the amount of memory needed to hold a device view of the specified column...
column_device_view(column_view column, void *h_ptr, void *d_ptr)
Creates an instance of this class using the specified host memory pointer (h_ptr) to store child obje...
const_iterator< T > end() const
Returns an iterator to the element following the last element of the column.
device_span< column_device_view const > children() const noexcept
Returns a span containing the children of this column.
A non-owning, immutable view of device data as a column of elements, some of which may be null as ind...
A container of nullable device data as a column of elements.
Definition: column.hpp:36
Indicator for the logical data type of an element in a column.
Definition: types.hpp:278
constexpr CUDF_HOST_DEVICE type_id id() const noexcept
Returns the type identifier.
Definition: types.hpp:322
CUDF_HOST_DEVICE data_type type() const noexcept
Returns the element type.
bool is_valid_nocheck(size_type element_index) const noexcept
Returns whether the specified element holds a valid value (i.e., not null)
CUDF_HOST_DEVICE bool nullable() const noexcept
Indicates whether the column can contain null elements, i.e., if it has an allocated bitmask.
A mutable, non-owning view of device data as a column of elements that is trivially copyable and usab...
A mutable, non-owning view of device data as a column of elements that is trivially copyable and usab...
static constexpr CUDF_HOST_DEVICE bool has_element_accessor()
For a given T, indicates if mutable_column_device_view::element<T>() has a valid overload.
static auto create(data_type type, size_type size, void const *data, bitmask_type const *null_mask, size_type offset, mutable_column_device_view *children, size_type num_children)
Factory to construct a mutable column view that is usable in device memory from pre-existing device m...
void destroy()
Destroy the mutable_column_device_view object.
mutable_column_device_view(mutable_column_view column, void *h_ptr, void *d_ptr)
Creates an instance of this class using the specified host memory pointer (h_ptr) to store child obje...
static std::size_t extent(mutable_column_view source_view)
Return the size in bytes of the amount of memory needed to hold a device view of the specified column...
mutable_column_device_view child(size_type child_index) const noexcept
Returns the specified child.
mutable_column_device_view(mutable_column_device_view const &)=default
Copy constructor.
T & element(size_type element_index) const noexcept
Returns reference to element at the specified index.
iterator< T > end()
Return one past the last element after underlying data is casted to the specified type.
mutable_column_device_view & operator=(mutable_column_device_view &&)=default
Move assignment operator.
mutable_column_device_view(mutable_column_device_view &&)=default
Move constructor.
cuda::counting_iterator< size_type > count_it
Counting iterator.
mutable_column_device_view & operator=(mutable_column_device_view const &)=default
Copy assignment operator.
thrust::transform_iterator< detail::mutable_value_accessor< T >, count_it > iterator
Iterator for navigating this column.
static std::unique_ptr< mutable_column_device_view, std::function< void(mutable_column_device_view *)> > create(mutable_column_view source_view, rmm::cuda_stream_view stream=cudf::get_default_stream(), rmm::device_async_resource_ref mr=cudf::get_current_device_resource_ref())
Factory to construct a column view that is usable in device memory.
iterator< T > begin()
Return first element (accounting for offset) after underlying data is casted to the specified type.
A non-owning, mutable view of device data as a column of elements, some of which may be null as indic...
ColumnDeviceView * child_columns_to_device_array(ColumnViewIterator child_begin, ColumnViewIterator child_end, void *h_ptr, void *d_ptr)
Helper function for use by column_device_view and mutable_column_device_view constructors to build de...
Column device view class definitions.
column view class definitions
APIs for querying the default CUDA stream and per-thread default stream status.
size_type null_count(bitmask_type const *bitmask, size_type start, size_type stop, rmm::cuda_stream_view stream=cudf::get_default_stream())
Given a validity bitmask, counts the number of null elements (unset bits) in the range [start,...
rmm::cuda_stream_view const get_default_stream()
Get the current default stream.
rmm::device_async_resource_ref get_current_device_resource_ref()
Get the current device memory resource reference.
cuda::mr::resource_ref< cuda::mr::device_accessible > device_async_resource_ref
CUDF_HOST_DEVICE constexpr decltype(auto) __forceinline__ type_dispatcher(cudf::data_type dtype, Functor f, Ts &&... args)
Invokes an operator() template with the type instantiation based on the specified cudf::data_type's i...
std::conditional_t< std::is_same_v< numeric::decimal32, T >, int32_t, std::conditional_t< std::is_same_v< numeric::decimal64, T >, int64_t, std::conditional_t< std::is_same_v< numeric::decimal128, T >, __int128_t, T > >> device_storage_type_t
"Returns" the corresponding type that is stored on the device when using cudf::column
#define CUDF_EXPECTS(...)
Macro for checking (pre-)conditions that throws an exception when a condition is violated.
Definition: error.hpp:182
cuda::std::span< T, Extent > device_span
Device span is an alias of cuda::std::span.
Definition: span.hpp:300
int32_t size_type
Row index type for columns and tables.
Definition: types.hpp:76
uint32_t bitmask_type
Bitmask type stored as 32-bit unsigned integer.
Definition: types.hpp:77
size_type distance(T f, T l)
Similar to std::distance but returns cudf::size_type and performs static_cast
Definition: types.hpp:91
#define CUDF_ENABLE_IF(...)
Convenience macro for SFINAE as an unnamed template parameter.
Definition: traits.hpp:43
Class definition for cudf::list_view.
APIs for getting and setting the current device memory resource.
cuDF interfaces
Definition: host_udf.hpp:26
bool has_nulls(table_view const &view)
Returns True if the table has nulls in any of its columns.
requires(is_index_type< IndexType >() &&is_relationally_comparable< KeyType, KeyType >()) struct dictionary_element
A type tag to specify that a column should be treated as a dictionary column.
APIs for spans.
Class definition for cudf::strings_column_view.
Class definition for cudf::struct_view.
Mutable value accessor of column without null bitmask.
T & operator()(cudf::size_type i)
Accessor.
mutable_value_accessor(mutable_column_device_view &_col)
Constructor.
mutable_column_device_view col
mutable column view of column in device
optional accessor of a column
cuda::std::optional< T > operator()(cudf::size_type i) const
Returns a cuda::std::optional of column[i].
column_device_view const col
column view of column in device
optional_accessor(column_device_view const &_col, Nullate with_nulls)
Constructor.
pair accessor of column with/without null bitmask
cuda::std::pair< T, bool > operator()(cudf::size_type i) const
Pair accessor.
column_device_view const col
column view of column in device
pair_accessor(column_device_view const &_col)
constructor
pair accessor of column with/without null bitmask
cuda::std::pair< rep_type, bool > operator()(cudf::size_type i) const
Pair accessor.
column_device_view const col
column view of column in device
device_storage_type_t< T > rep_type
representation type
pair_rep_accessor(column_device_view const &_col)
constructor
value accessor of column without null bitmask
column_device_view const col
column view of column in device
value_accessor(column_device_view const &_col)
constructor
T operator()(cudf::size_type i) const
Returns the value of element at index i
A strongly typed wrapper for indices in a DICTIONARY type column.
Definition: dictionary.hpp:39
Defines the mapping between cudf::type_id runtime type information and concrete C++ types.
#define CUDF_HOST_DEVICE
Indicates that the function or method is usable on host and device.
Definition: types.hpp:21